@@ -1,6 +1,6 @@
 /* Contains the basic I/O and setup routines.  */
 
-/*  Copyright (C) 2001  Susan Bassein
+/*  Copyright (C) 2001 Free Software Foundation, Inc.
  *
  *  Dap is a GNU program.
  *
@@ -766,6 +766,7 @@ int v;				/* index to variables */
 int l;
 int i;
 int iv;
+int nread;	/* number of chars read by getline */
 
 if (!stepinit)
 	{
@@ -798,15 +799,19 @@ for (iv = 0; iv < dap_obs[dap_ono].do_ivar; iv++)
 		}
 	}
 dap_prev[dap_ono].do_valid = dap_obs[dap_ono].do_valid;
-if (getline(dap_in[dap_ono], line) < 0)
+if ((nread = getline(dap_in[dap_ono], line)) <= 0)
 	{
 	eof[dap_ono] = 1;
 	dap_obs[dap_ono].do_valid = 0;
 	fprintf(stderr, "(step) %d lines read from  %s\n",
 		lineno[dap_ono] - 1, dap_in[dap_ono]->dfile_name);
+	if (!nread)
+		fputs("(step) WARNING: terminated on null line\n", stderr);
 	fflush(stderr);
 	fprintf(dap_log, "(step) %d lines read from  %s\n",
 		lineno[dap_ono] - 1, dap_in[dap_ono]->dfile_name);
+	if (!nread)
+		fputs("(step) WARNING: terminated on null line\n", dap_log);
 	return 0;
 	}
 if (nfields && dap_obs[dap_ono].do_ivar != nfields)
@@ -826,15 +831,17 @@ for (v = 0, l = 0; v < dap_obs[dap_ono].do_ivar; v++)
 		if (i < fieldwd[v])
 			{
 			fprintf(dap_log,
-				"(step (%s:%d)) ERROR: incomplete fixed length field for %s: %s\n",
+		"(step (%s:%d)) ERROR: got %d of %d characters for fixed length field for %s: %s\n",
 				dap_in[dap_ono]->dfile_name,
-				lineno[dap_ono],
+				lineno[dap_ono], i, fieldwd[v],
 				dap_obs[dap_ono].do_nam[dap_obs[dap_ono].do_in[v]],
 				value);
 			fprintf(dap_err,
-				"(step (%s:%d)) ERROR: got %d of %d characters.\n",
-					dap_in[dap_ono]->dfile_name,
-					lineno[dap_ono], i, fieldwd[v]);
+		"(step (%s:%d)) ERROR: got %d of %d characters for fixed length field for %s: %s\n",
+				dap_in[dap_ono]->dfile_name,
+				lineno[dap_ono], i, fieldwd[v],
+				dap_obs[dap_ono].do_nam[dap_obs[dap_ono].do_in[v]],
+				value);
 			exit(1);
 			}
 		}
@@ -982,14 +989,14 @@ while (varspec[s])
 			{
 			if (dap_obs[dap_ono].do_len[v] != vlen)
 				{
-				if (dap_obs[dap_ono].do_len[v] > 0)
+				if (vlen > 0 && dap_obs[dap_ono].do_len[v] > 0)
 					fprintf(dap_err,
 				"(dap_vd) respecification of length of %s from %d to %d\n",
 						varnam, dap_obs[dap_ono].do_len[v] + 1, vlen + 1);
 				else
 					fprintf(dap_err,
-				"(dap_vd) respecification of length of %s from %d to %d\n",
-						varnam, dap_obs[dap_ono].do_len[v], vlen);
+					"(dap_vd) respecification of type of %s\n",
+						varnam);
 				exit(1);
 				}
 			}
@@ -1013,6 +1020,12 @@ while (varspec[s])
 			if (dap_ono < 2)
 				dap_prev[dap_ono].do_str[v] = dap_malloc(vlen + 1, varnam);
 			}
+		if (vlen == DBL)
+			dap_obs[dap_ono].do_dbl[v] = 0.0 / 0.0;
+		else if (vlen == INT)
+			dap_obs[dap_ono].do_int[v] = 0;
+		else
+			dap_obs[dap_ono].do_str[v][0] = '\0';
 		s += i;
 		while (varspec[s] == ' ')
 			s++;
@@ -1036,13 +1049,19 @@ int dim;
 int d;
 
 if ((v = dap_varnum(varname)) >= 0)
+	{
 	dap_obs[dap_ono].do_dl[v] = dbl;
+	*dbl = 0.0 / 0.0;
+	}
 else
 	{
 	if ((v = dap_arrnum(varname, &dim)) > 0)
 		{
 		for (d = 0; d < dim; d++)
+			{
 			dap_obs[dap_ono].do_dl[v + d] = dbl + d;
+			dbl[d] = 0.0 / 0.0;
+			}
 		}
 	else
 		{
@@ -1059,13 +1078,19 @@ int dim;
 int d;
 
 if ((v = dap_varnum(varname)) > 0)
+	{
 	dap_obs[dap_ono].do_il[v] = i;
+	*i = 0;
+	}
 else
 	{
 	if ((v = dap_arrnum(varname, &dim)) > 0)
 		{
 		for (d = 0; d < dim; d++)
+			{
 			dap_obs[dap_ono].do_il[v + d] = i + d;
+			i[d] = 0;
+			}
 		}
 	else
 		{
@@ -1084,6 +1109,7 @@ if ((v = dap_varnum(varname)) >= 0)
 	dap_free(dap_obs[dap_ono].do_str[v]);
 	dap_obs[dap_ono].do_str[v] = s;
 	dap_obs[dap_ono].do_sl[v] = 1;
+	s[0] = '\0';
 	}
 else
 	{
