@@ -1,6 +1,6 @@
 /* dap1.c -- sort and table */
 
-/*  Copyright (C) 2001  Susan Bassein
+/*  Copyright (C) 2001 Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -27,7 +27,8 @@ extern DFILE *dap_out;		/* output dataset file */
 extern dataobs dap_obs[];	/* values from current input line */
 extern dataobs dap_prev[];	/* values from previous input line */
 extern FILE *dap_lst;		/* .lst file */
-extern FILE *dap_err;		/* .log file */
+extern FILE *dap_err;		/* .err file */
+extern FILE *dap_log;		/* .log file */
 extern char *dap_title;		/* title for sections in .lst file */
 
 extern int dap_maxlines;	/* maximum number of lines in a ramfile */
@@ -418,6 +419,7 @@ int vn;
 int wn;
 int nvar;
 int s;
+int nonly;	/* only requested N, for var that didn't exist */
 
 if (!varlist)
 	{
@@ -433,7 +435,7 @@ nvar = 0;
 tmplist = dap_malloc(dap_listlen + 1, "");
 tmplist[0] = '\0';
 wn = -1;
-while (v >= 0)
+for (nonly = 0; v >= 0; )
 	{
 	for (i = v; i >= 0 && varlist[i] != ' ' && varlist[i] != '*'; --i)
 		;
@@ -481,6 +483,7 @@ while (v >= 0)
 		strcpy(tmplist, vname);
 		strcat(vname, " -1");
 		vn = dap_vd(vname, 0);
+		nonly = 1;
 		}
 	v = i;
 	if (v >= 0 && varlist[v] == '*')
@@ -514,6 +517,8 @@ for (outlist[0] = '\0'; i >= 0; )
 	}
 dap_free(vname);
 dap_free(tmplist);
+if (nonly)
+	return -nvar;
 return nvar;
 }
 
@@ -638,10 +643,10 @@ for (pctptn = 0; statlist[s]; )
 dap_free(stat);
 }
 
-static void meansout(int varv[], int nvar, int nobs, double sum[], double sumwt[],
+static void meansout(int varv[], int nvar, int nobs[], double sum[], double sumwt[],
 	double min[], double max[], double ss[], int stats[])
 {
-double dn;
+double *dn;
 int typevar;
 int v;
 int nsteps;
@@ -650,214 +655,262 @@ double *range;
 double fract;
 
 dap_swap();
-dn = (double) nobs;
+dn = (double *) dap_malloc(sizeof(double) * nvar, "");
+for (v = 0; v < nvar; v++)
+	dn[v] = (double) nobs[v];
 if ((typevar = dap_varnum("_type_")) < 0)
 	{
 	fprintf(dap_err, "(meansout) Missing _type_ variable\n");
 	exit(1);
 	}
-if (nobs >= 1)
+if (stats[N])
 	{
-	if (stats[N])
+	strcpy(dap_obs[0].do_str[typevar], "N");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "N");
-		for (v = 0; v < nvar; v++)
-			dap_obs[0].do_dbl[varv[v]] = dn;
-		output();
+		if (nobs[v] >= 1)
+			dap_obs[0].do_dbl[varv[v]] = dn[v];
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[SUM])
+	output();
+	}
+if (stats[SUM])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SUM");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SUM");
-		for (v = 0; v < nvar; v++)
+		if (nobs[v] >= 1)
 			dap_obs[0].do_dbl[varv[v]] = sum[v];
-		output();
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[SUMWT])
+	output();
+	}
+if (stats[SUMWT])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SUMWT");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SUMWT");
-		for (v = 0; v < nvar; v++)
+		if (nobs[v] >= 1)
 			dap_obs[0].do_dbl[varv[v]] = sumwt[v];
-		output();
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[MEAN])
+	output();
+	}
+if (stats[MEAN])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "MEAN");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "MEAN");
-		for (v = 0; v < nvar; v++)
+		if (nobs[v] >= 1)
 			dap_obs[0].do_dbl[varv[v]] = sum[v] / sumwt[v];
-		output();
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[MIN])
+	output();
+	}
+if (stats[MIN])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "MIN");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "MIN");
-		for (v = 0; v < nvar; v++)
+		if (nobs[v] >= 1)
 			dap_obs[0].do_dbl[varv[v]] = min[v];
-		output();
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[MAX])
+	output();
+	}
+if (stats[MAX])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "MAX");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "MAX");
-		for (v = 0; v < nvar; v++)
+		if (nobs[v] >= 1)
 			dap_obs[0].do_dbl[varv[v]] = max[v];
-		output();
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[RANGE])
+	output();
+	}
+if (stats[RANGE])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "RANGE");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "RANGE");
-		for (v = 0; v < nvar; v++)
+		if (nobs[v] >= 1)
 			dap_obs[0].do_dbl[varv[v]] = max[v] - min[v];
-		output();
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[VAR])
+	output();
+	}
+if (stats[VAR])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "VAR");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "VAR");
-		for (v = 0; v < nvar; v++)
-			{
-			if (dn > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = ss[v] / (dn - 1.0);
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (dn[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] = ss[v] / (dn[v] - 1.0);
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[SD])
+	output();
+	}
+if (stats[SD])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SD");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SD");
-		for (v = 0; v < nvar; v++)
-			{
-			if (dn > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = sqrt(ss[v] / (dn - 1.0));
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (dn[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+					sqrt(ss[v] / (dn[v] - 1.0));
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[SEM])
+	output();
+	}
+if (stats[SEM])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SEM");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SEM");
-		for (v = 0; v < nvar; v++)
-			{
-			if (dn > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = sqrt(ss[v] / (dn * (dn - 1.0)));
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (dn[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+					sqrt(ss[v] / (dn[v] * (dn[v] - 1.0)));
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[VARM])
+	output();
+	}
+if (stats[VARM])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "VARM");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "VARM");
-		for (v = 0; v < nvar; v++)
-			{
-			if (dn > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = ss[v] / (dn * (dn - 1.0));
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (dn[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+					ss[v] / (dn[v] * (dn[v] - 1.0));
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[VARFREQ])
+	output();
+	}
+if (stats[VARFREQ])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "VARFREQ");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "VARFREQ");
-		for (v = 0; v < nvar; v++)
-			{
-			if (sumwt[v] > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = ss[v] / (sumwt[v] - 1.0);
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (sumwt[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+					ss[v] / (sumwt[v] - 1.0);
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[SDFREQ])
+	output();
+	}
+if (stats[SDFREQ])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SDFREQ");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SDFREQ");
-		for (v = 0; v < nvar; v++)
-			{
-			if (sumwt[v] > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = sqrt(ss[v] / (sumwt[v] - 1.0));
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (sumwt[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+					sqrt(ss[v] / (sumwt[v] - 1.0));
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[SEMFREQ])
+	output();
+	}
+if (stats[SEMFREQ])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SEMFREQ");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SEMFREQ");
-		for (v = 0; v < nvar; v++)
-			{
-			if (sumwt[v] > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = sqrt(ss[v] / (sumwt[v] * (sumwt[v] - 1.0)));
-			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
-			}
-		output();
+		if (sumwt[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+				sqrt(ss[v] / (sumwt[v] * (sumwt[v] - 1.0)));
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[VARMFREQ])
+	output();
+	}
+if (stats[VARMFREQ])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "SEMFREQ");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "SEMFREQ");
-		for (v = 0; v < nvar; v++)
+		if (sumwt[v] > 1.0)
+			dap_obs[0].do_dbl[varv[v]] =
+				ss[v] / (sumwt[v] * (sumwt[v] - 1.0));
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
+		}
+	output();
+	}
+if (stats[STEP])
+	{
+	if (sscanf(dap_sttnm[STEP] + 4, "%d", &nsteps) == 1)
+		{
+		range = (double *) dap_malloc(sizeof(double) * nvar, "");
+		strcpy(dap_obs[0].do_str[typevar], "STEP");
+		for (v = 0; v < nvar; v++) 
 			{
-			if (sumwt[v] > 1.0)
-				dap_obs[0].do_dbl[varv[v]] = ss[v] / (sumwt[v] * (sumwt[v] - 1.0));
+			if (nobs[v] >= 1)
+				range[v] = max[v] - min[v];
 			else
-				dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
+				range[v] = 0.0 / 0.0;
 			}
-		output();
+		for (step = 0; step <= nsteps; step++)
+			{
+			fract = ((double) step) / ((double) nsteps);
+			for (v = 0; v < nvar; v++)
+				dap_obs[0].do_dbl[varv[v]] =
+					min[v] + range[v] * fract;
+			output();
+			}
+		dap_free(range);
+		}
+	else
+		{
+		fprintf(dap_err, "(meansout) Bad number of steps: %s\n",
+			dap_sttnm[STEP] + 4);
+		exit(1);
 		}
 	}
-if (nobs > 1)
+if (stats[T])
 	{
-	if (stats[STEP])
+	strcpy(dap_obs[0].do_str[typevar], "T");
+	for (v = 0; v < nvar; v++)
 		{
-		if (sscanf(dap_sttnm[STEP] + 4, "%d", &nsteps) == 1)
-			{
-			range = (double *) dap_malloc(sizeof(double) * nvar, "");
-			strcpy(dap_obs[0].do_str[typevar], "STEP");
-			for (v = 0; v < nvar; v++)
-				range[v] = (max[v] - min[v]);
-			for (step = 0; step <= nsteps; step++)
-				{
-				fract = ((double) step) / ((double) nsteps);
-				for (v = 0; v < nvar; v++)
-					dap_obs[0].do_dbl[varv[v]] =
-							min[v] + range[v] * fract;
-				output();
-				}
-			dap_free(range);
-			}
+		if (sumwt[v] > 0.0 && ss[v] > 0.0)
+			dap_obs[0].do_dbl[varv[v]] = (sum[v] /
+					sumwt[v]) * sqrt(sumwt[v] *
+					(dn[v] - 1.0) / ss[v]);
 		else
-			{
-			fprintf(dap_err, "(meansout) Bad number of steps: %s\n",
-				dap_sttnm[STEP] + 4);
-			exit(1);
-			}
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
-	if (stats[T])
+	output();
+	}
+if (stats[TPROB])
+	{
+	strcpy(dap_obs[0].do_str[typevar], "TPROB");
+	for (v = 0; v < nvar; v++)
 		{
-		strcpy(dap_obs[0].do_str[typevar], "T");
-		for (v = 0; v < nvar; v++)
-			{
-			if (sumwt[v] > 0.0 && ss[v] > 0.0)
-				dap_obs[0].do_dbl[varv[v]] = (sum[v] /
-						sumwt[v]) * sqrt(sumwt[v] *
-						(dn - 1.0) / ss[v]);
-			}
-		output();
-		}
-	if (stats[TPROB])
-		{
-		strcpy(dap_obs[0].do_str[typevar], "TPROB");
-		for (v = 0; v < nvar; v++)
-			{
-			if (sumwt[v] > 0.0 && ss[v] > 0.0)
-				dap_obs[0].do_dbl[varv[v]] =
-						2.0 * probt(fabs((sum[v] /
-						sumwt[v]) * sqrt(sumwt[v] *
-						(dn - 1.0) / ss[v])), nobs - 1);
-			}
-		output();
+		if (sumwt[v] > 0.0 && ss[v] > 0.0)
+			dap_obs[0].do_dbl[varv[v]] =
+					2.0 * probt(fabs((sum[v] /
+					sumwt[v]) * sqrt(sumwt[v] *
+					(dn[v] - 1.0) / ss[v])), nobs[v] - 1);
+		else
+			dap_obs[0].do_dbl[varv[v]] = 0.0 / 0.0;
 		}
+	output();
 	}
 dap_swap();
+dap_free(dn);
 }
 
 int dap_list(char *varlist, int *varv, int maxvars)
@@ -952,11 +1005,12 @@ void means(char *fname, char *varlist, char *statlist, char *marks)
 {
 char *outname;
 int stats[NSTATS];
+int nonly;
 int *varv;
 int *markv;
 int nvar;
 int nmark;
-int nobs;
+int *nobs;
 char *outlist;
 int *wtvar;
 double *sum;
@@ -968,6 +1022,7 @@ int v;
 double wt;
 double vtmp;
 double tmp;
+int *nnan;		/* number of NaN's for each variable */
 int more;
 
 if (!fname)
@@ -984,6 +1039,10 @@ outlist = dap_malloc(dap_listlen + 1, "");
 inset(fname);
 dap_stats(statlist, stats);
 nvar = dap_mnsparse(varlist, outlist, varv, wtvar, stats);
+if ((nonly = (nvar < 0)))
+	nvar = -nvar;
+nobs = (int *) dap_malloc(sizeof(int) * nvar, "");
+nnan = (int *) dap_malloc(sizeof(int) * nvar, "");
 if (marks && marks[0])
 	{
 	strcat(outlist, " ");
@@ -1001,8 +1060,10 @@ for (v = 0; v < nvar; v++)
 	sum[v] = 0.0;
 	sumwt[v] = 0.0;
 	ss[v] = 0.0;
+	nobs[v] = 0;
+	nnan[v] = 0;
 	}
-for (nobs = 0, more = 1; more; nobs++)
+for (more = 1; more; )
 	{
 	more = step();
 	if (dap_newpart(markv, nmark))
@@ -1010,11 +1071,19 @@ for (nobs = 0, more = 1; more; nobs++)
 		meansout(varv, nvar, nobs, sum, sumwt, min, max, ss, stats);
 		for (v = 0; v < nvar; v++)
 			{
+			if (nnan[v])
+				{
+				dap_swap();
+				fprintf(dap_log, "means: %d NaNs for %s\n",
+					nnan[v], dap_obs[0].do_nam[varv[v]]);
+				dap_swap();
+				}
 			sum[v] = 0.0;
 			sumwt[v] = 0.0;
 			ss[v] = 0.0;
+			nobs[v] = 0;
+			nnan[v] = 0;
 			}
-		nobs = 0;
 		}
 	for (v = 0; v < nvar; v++)
 		{
@@ -1023,27 +1092,36 @@ for (nobs = 0, more = 1; more; nobs++)
 			wt = dap_obs[0].do_dbl[wtvar[v]];
 		else
 			wt = 1.0;
-		if (!nobs)
+		if (finite(vtmp) && finite(wt))
 			{
-			min[v] = vtmp;
-			max[v] = vtmp;
-			}
-		else
-			{
-			if (vtmp < min[v])
+			if (!nobs[v])
+				{
 				min[v] = vtmp;
-			if (vtmp > max[v])
 				max[v] = vtmp;
-			tmp = sum[v] - sumwt[v] * vtmp;
-			ss[v] += tmp * tmp * wt / (sumwt[v] * (sumwt[v] + wt));
+				}
+			else
+				{
+				if (vtmp < min[v])
+					min[v] = vtmp;
+				if (vtmp > max[v])
+					max[v] = vtmp;
+				tmp = sum[v] - sumwt[v] * vtmp;
+				ss[v] += tmp * tmp * wt / (sumwt[v] * (sumwt[v] + wt));
+				}
+			sumwt[v] += wt;
+			sum[v] += vtmp * wt;
+			nobs[v]++;
 			}
-		sumwt[v] += wt;
-		sum[v] += vtmp * wt;
+		else if (nonly)
+			nobs[v]++;
+		else
+			nnan[v]++;
 		}
 	}
 dap_free(outname);
 dap_free(varv);
 dap_free(markv);
+dap_free(nobs);
 dap_free(outlist);
 dap_free(wtvar);
 dap_free(sum);
@@ -1051,6 +1129,7 @@ dap_free(sumwt);
 dap_free(ss);
 dap_free(min);
 dap_free(max);
+dap_free(nnan);
 }
 
 typedef struct
