@@ -1,6 +1,6 @@
 /* dap2.c -- statistics */
 
-/*  Copyright (C) 2001  Susan Bassein
+/*  Copyright (C) 2001 Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -291,6 +291,13 @@ if (stats[MAX])
 	strcpy(dap_obs[0].do_str[typen], "MAX");
 	output();
 	}
+if (stats[RANGE])
+	{
+	for (v = 0; v < nvar; v++)
+		dap_obs[0].do_dbl[varv[v]] = val[v][nobs - 1][0] - val[v][0][0];
+	strcpy(dap_obs[0].do_str[typen], "RANGE");
+	output();
+	}
 if (stats[QRANGE])
 	{
 	for (v = 0; v < nvar; v++)
@@ -387,6 +394,13 @@ for (nobs = 0, more = 1; more; nobs++)
 				val[v][nobs][1] = dap_obs[0].do_dbl[wtvar[v]];
 			else
 				val[v][nobs][1] = 1.0;
+			if (!finite(val[v][nobs][0]) || !finite(val[v][nobs][1]))
+				{
+				fprintf(dap_err,
+					"(pctiles) NaN value %d for %s\n",
+					nobs, dap_obs[0].do_nam[varv[v]]);
+				exit(1);
+				}
 			}
 		}
 	else
@@ -427,66 +441,48 @@ for (v = 0; v < nvar; v++)
 strcpy(dap_obs[0].do_str[typen], "N");
 for (v = 0; v < nvar; v++)
 	{
-	if (dap_obs[0].do_len[varv[v]] == DBL)
+	strcpy(dap_obs[0].do_str[varn[0]], dap_obs[0].do_nam[varv[v]]);
+	for (w = 0; w < nvar; w++)
 		{
-		strcpy(dap_obs[0].do_str[varn[0]], dap_obs[0].do_nam[varv[v]]);
-		for (w = 0; w < nvar; w++)
-			{
-			if (dap_obs[0].do_len[varv[w]] == DBL)
-				{
-				strcpy(dap_obs[0].do_str[varn[1]],
-						dap_obs[0].do_nam[varv[w]]);
-				dap_obs[0].do_dbl[varn[2]] = (double) nobs;
-				output();
-				}
-			}
+		strcpy(dap_obs[0].do_str[varn[1]],
+				dap_obs[0].do_nam[varv[w]]);
+		dap_obs[0].do_dbl[varn[2]] = (double) nobs;
+		output();
 		}
 	}
 strcpy(dap_obs[0].do_str[typen], "CORR");
 for (v = 0; v < nvar; v++)
 	{
-	if (dap_obs[0].do_len[varv[v]] == DBL)
+	strcpy(dap_obs[0].do_str[varn[0]], dap_obs[0].do_nam[varv[v]]);
+	for (w = 0; w < nvar; w++)
 		{
-		strcpy(dap_obs[0].do_str[varn[0]], dap_obs[0].do_nam[varv[v]]);
-		for (w = 0; w < nvar; w++)
-			{
-			if (dap_obs[0].do_len[varv[w]] == DBL)
-				{
-				strcpy(dap_obs[0].do_str[varn[1]],
-						dap_obs[0].do_nam[varv[w]]);
-				if (w == v)
-					cormat[v][w] = 1.0;
-				else
-					cormat[v][w] /= sqrt(ss[v] * ss[w]);
-				dap_obs[0].do_dbl[varn[2]] = cormat[v][w];
-				output();
-				}
-			}
+		strcpy(dap_obs[0].do_str[varn[1]],
+				dap_obs[0].do_nam[varv[w]]);
+		if (w == v)
+			cormat[v][w] = 1.0;
+		else
+			cormat[v][w] /= sqrt(ss[v] * ss[w]);
+		dap_obs[0].do_dbl[varn[2]] = cormat[v][w];
+		output();
 		}
 	}
 strcpy(dap_obs[0].do_str[typen], "PCORR");
 for (v = 0; v < nvar; v++)
         {
-        if (dap_obs[0].do_len[varv[v]] == DBL)
-                {
-		strcpy(dap_obs[0].do_str[varn[0]], dap_obs[0].do_nam[varv[v]]);
-                for (w = 0; w < nvar; w++)
-                        {
-                        if (dap_obs[0].do_len[varv[w]] == DBL)
-				{
-				strcpy(dap_obs[0].do_str[varn[1]],
-						dap_obs[0].do_nam[varv[w]]);
-                                r = fabs(cormat[v][w]);
-				if (r == 1.0)
-					dap_obs[0].do_dbl[varn[2]] = 0.0;
-				else
-					dap_obs[0].do_dbl[varn[2]] =
-						2.0 * probt(nf * r / sqrt(1.0 - r * r),
-								nobs - 2);
-				output();
-				}
-                        }
-                }
+	strcpy(dap_obs[0].do_str[varn[0]], dap_obs[0].do_nam[varv[v]]);
+	for (w = 0; w < nvar; w++)
+		{
+		strcpy(dap_obs[0].do_str[varn[1]],
+				dap_obs[0].do_nam[varv[w]]);
+		r = fabs(cormat[v][w]);
+		if (r == 1.0)
+			dap_obs[0].do_dbl[varn[2]] = 0.0;
+		else
+			dap_obs[0].do_dbl[varn[2]] =
+				2.0 * probt(nf * r / sqrt(1.0 - r * r),
+						nobs - 2);
+		output();
+		}
         }
 dap_swap();
 }
@@ -550,6 +546,12 @@ outset(outname, outlist);
 nmark = dap_list(marks, markv, dap_maxvar);
 for (v = 0; v < nvar; v++)
 	{
+	if (dap_obs[0].do_len[varv[v]] != DBL)
+		{
+		fprintf(dap_err, "(corr) variables must be of type double: %s\n",
+			dap_obs[0].do_nam[varv[v]]);
+		exit(1);
+		}
 	for (w = v + 1; w < nvar; w++)
 		cormat[v][w] = 0.0;
 	sum[v] = 0.0;
@@ -575,27 +577,26 @@ for (nobs = 0, more = 1; more; nobs++)
 		dn = (double) nobs;
 		for (v = 0; v < nvar; v++)
 			{
-			if (dap_obs[0].do_len[varv[v]] == DBL)
+			vtmp = dap_obs[0].do_dbl[varv[v]];
+			if (!finite(vtmp))
 				{
-				vtmp = dap_obs[0].do_dbl[varv[v]];
-				if (nobs)
-					{
-					tmp = sum[v] - dn * vtmp;
-					for (w = v + 1; w < nvar; w++)
-						{
-						if (dap_obs[0].do_len[varv[w]] == DBL)
-							{
-							cormat[v][w] +=
-								tmp *
-								(sum[w] - dn *
-								dap_obs[0].do_dbl[varv[w]]) /
-								(dn * (dn + 1.0));
-							}
-						}
-					ss[v] += tmp * tmp / (dn * (dn + 1.0));
-					}
-				sum[v] += vtmp;
+				fprintf(dap_err,
+					"(corr) NaN value %d for %s\n",
+					nobs, dap_obs[0].do_nam[varv[v]]);
+				exit(1);
 				}
+			if (nobs)
+				{
+				tmp = sum[v] - dn * vtmp;
+				for (w = v + 1; w < nvar; w++)
+					{
+					cormat[v][w] += tmp * (sum[w] - dn *
+						dap_obs[0].do_dbl[varv[w]]) /
+						(dn * (dn + 1.0));
+					}
+				ss[v] += tmp * tmp / (dn * (dn + 1.0));
+				}
+			sum[v] += vtmp;
 			}
 		}
 	}
@@ -643,7 +644,7 @@ for (nvar = 0, number = 0; varspec[s]; nvar++)
 		else
 			{
 			varname[i] = '\0';
-			fprintf(dap_err, "(groupparse) Variable name too long: %s\n",
+			fprintf(dap_err, "(groupparse) variable name too long: %s\n",
 					varname);
 			exit(1);
 			}
@@ -663,9 +664,10 @@ for (nvar = 0, number = 0; varspec[s]; nvar++)
 					'0' <= varspec[s + i] && varspec[s + i] <= '9';
 						i++)
 					n = 10 * n + varspec[s + i] - '0';
-				if (varspec[s + i] != ' ')
+				if (!n || varspec[s + i] != ' ')
 					{
-					fprintf(dap_err, "(groupparse) Invalid ngroups: %s\n",
+					fprintf(dap_err,
+				"(groupparse) invalid number of groups: %s\n",
 						varspec + s);
 					exit(1);
 					}
@@ -680,7 +682,7 @@ for (nvar = 0, number = 0; varspec[s]; nvar++)
 						{
 						ctype[i] = '\0';
 						fprintf(dap_err,
-						"(groupparse) Class type too long: %s\n",
+						"(groupparse) class type too long: %s\n",
 								ctype);
 						exit(1);
 						}
@@ -696,28 +698,31 @@ for (nvar = 0, number = 0; varspec[s]; nvar++)
 				else
 					{
 					fprintf(dap_err,
-						"(groupparse) Invalid class type: %s\n",
+						"(groupparse) invalid class type: %s\n",
 						ctype);
 					exit(1);
 					}
-				if (classtype[nvar])
+				if (n > dap_maxbars)
 					{
-					if (n > dap_maxbars)
-						{
-						fprintf(dap_err,
-						"(groupparse) Too many classes: %d\n", n);
-						exit(1);
-						}
+					fprintf(dap_err,
+					"(groupparse) too many classes: %d\n", n);
+					exit(1);
 					}
 				}
 			else
 				{
 				fprintf(dap_err,
-					"(groupparse) Grouping variable must be dapdbl: %s\n",
+				"(groupparse) grouping variable must be of type double: %s\n",
 					varname);
 				exit(1);
 				}
 			}
+		else
+			{
+			fprintf(dap_err, "(groupparse) missing number of groups for %s\n",
+					varname);
+			exit(1);
+			}
 		}
 	else if (!nvar)
 		{
@@ -738,20 +743,20 @@ for (nvar = 0, number = 0; varspec[s]; nvar++)
 		}
 	else
 		{
-		fprintf(dap_err, "(groupparse) Unknown variable: %s\n", varname);
+		fprintf(dap_err, "(groupparse) unknown variable: %s\n", varname);
 		exit(1);
 		}
 	while (varspec[s] == ' ')
 		s++;
 	}
-if (number < 0)
-	return number;
 dap_free(varname);
 dap_free(ctype);
+if (number < 0)
+	return number;
 return nvar;
 }
 
-static void getpoints(double **numval, int *numv, int nonum,
+static void getpoints(double **numval, int nonum,
 				int *ctype, int nobs, double **point)
 {
 int v;
@@ -764,22 +769,22 @@ for (v = 0; v < nonum; v++)
 	{
 	qsort((void *) numval[v], (size_t) nobs,
 			(size_t) sizeof(double), cmp);
-	if (ctype[numv[v]] < 0)
+	if (ctype[v] < 0)
 		{
 		width = (numval[v][nobs - 1] - numval[v][0]) /
-				((double) -ctype[numv[v]]);
-		for (p = 0; p < -ctype[numv[v]]; p++)
+				((double) -ctype[v]);
+		for (p = 0; p < -ctype[v]; p++)
 			point[v][p] = numval[v][0] +
 				((double) p) * width;
 		point[v][p] = numval[v][nobs - 1];
 		}
-	else if (ctype[numv[v]] > 0)
+	else if (ctype[v] > 0)
 		{
-		for (p = 0; p < ctype[numv[v]]; p++)
+		for (p = 0; p < ctype[v]; p++)
 			{
 			index = (int) floor(((double) (p * nobs)) /
-						((double) ctype[numv[v]]));
-			excess = ((double) (p * nobs)) / ((double) ctype[numv[v]]) -
+						((double) ctype[v]));
+			excess = ((double) (p * nobs)) / ((double) ctype[v]) -
 					(double) index;
 			if (excess > 0.0 && index < nobs - 1)
 				point[v][p] = 0.5 *
@@ -787,7 +792,7 @@ for (v = 0; v < nonum; v++)
 			else
 				point[v][p] =
 				numval[v][(int) rint(((double) (p * nobs)) /
-							((double) ctype[numv[v]]))];
+							((double) ctype[v]))];
 			}
 		point[v][p] = numval[v][nobs - 1];
 		}
@@ -801,8 +806,6 @@ int *markv;
 int nmark;
 int *varv;
 int *ctype;
-int *numv;
-int nonum;
 double *nummem;
 double **numval;
 int nvar;
@@ -829,26 +832,16 @@ grpname = dap_malloc(dap_namelen + 3, "");
 markv = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
 varv = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
 ctype = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
-numv = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
 inset(fname);
 nmark = dap_list(marks, markv, dap_maxvar);
 nvar = groupparse(varspec, varv, ctype);
-nonum = 0;
 grpv = (int *) dap_malloc(sizeof(int) * (nvar ? nvar : 1), "");
 if (nvar > 0)
 	{
+	strcpy(grpname, "_");
 	for (v = 0; v < nvar; v++)
 		{
-		if (ctype[v])
-			{
-			ctype[nonum] = ctype[v];
-			numv[nonum++] = v;
-			}
-		}
-	strcpy(grpname, "_");
-	for (v = 0; v < nonum; v++)
-		{
-		strcpy(grpname + 1, dap_obs[0].do_nam[varv[numv[v]]]);
+		strcpy(grpname + 1, dap_obs[0].do_nam[varv[v]]);
 		grpname[dap_namelen] = '\0';
 		strcat(grpname, " -1");
 		grpv[v] = dap_vd(grpname, 0);
@@ -859,28 +852,24 @@ else
 outset(outname, "");
 nummem = NULL;
 numval = NULL;
-if (nonum)
+if (nvar > 0)
 	{
-	nummem = (double *) dap_malloc(sizeof(double) * nonum * dap_maxval, "");
-	numval = (double **) dap_malloc(nonum * dap_maxval * sizeof(double *), "");
-	for (v = 0; v < nonum; v++)
+	nummem = (double *) dap_malloc(sizeof(double) * nvar * dap_maxval, "");
+	numval = (double **) dap_malloc(nvar * dap_maxval * sizeof(double *), "");
+	ptmem = (double *) dap_malloc(sizeof(double) * nvar * (dap_maxbars + 1), "");
+	point = (double **) dap_malloc(sizeof(double *) * nvar, "");
+	for (v = 0; v < nvar; v++)
+		{
 		numval[v] = nummem + v * dap_maxval;
+		point[v] = ptmem + v * (dap_maxbars + 1);
+		}
 	}
-else if (nvar >= 0)
-	{
-	fprintf(dap_err, "(group) No grouping variables: %s\n", varspec);
-	exit(1);
-	}
-ptmem = (double *) dap_malloc(sizeof(double) * nonum * (dap_maxbars + 1), "");
-point = (double **) dap_malloc(sizeof(double *) * nonum, "");
-for (v = 0; v < nonum; v++)
-	point[v] = ptmem + v * (dap_maxbars + 1);
 for (dap_mark(), nobs = 0, more = 1; more; nobs++)
 	{
 	more = step();
 	if (dap_newpart(markv, nmark))
 		{
-		getpoints(numval, numv, nonum, ctype, nobs, point);
+		getpoints(numval, nvar, ctype, nobs, point);
 		dnobs = (double) nobs;
 		dap_rewind();
 		count = 1.0;
@@ -909,16 +898,13 @@ for (dap_mark(), nobs = 0, more = 1; more; nobs++)
 				dap_obs[0].do_dbl[grpv[0]] = 100.0 * count / dnobs;
 				break;
 			default:
-				for (v = 0; v < nonum; v++)
+				for (v = 0; v < nvar; v++)
 					{
-					if (ctype[v])
-						{
-						for (p = 1;
-							dap_obs[0].do_dbl[varv[numv[v]]] >
-								point[v][p]; p++)
-							;
-						dap_obs[0].do_dbl[grpv[v]] = (double) p;
-						}
+					for (p = 1;
+						dap_obs[0].do_dbl[varv[v]] >
+							point[v][p]; p++)
+						;
+					dap_obs[0].do_dbl[grpv[v]] = (double) p;
 					}
 				break;
 				}
@@ -927,24 +913,35 @@ for (dap_mark(), nobs = 0, more = 1; more; nobs++)
 			}
 		nobs = 0;
 		}
-	if (nonum)
+	if (nvar > 0)
 		{
 		if (nobs < dap_maxval)
 			{
-			for (v = 0; v < nonum; v++)
-				numval[v][nobs] = dap_obs[0].do_dbl[varv[numv[v]]];
+			for (v = 0; v < nvar; v++)
+				{
+				numval[v][nobs] = dap_obs[0].do_dbl[varv[v]];
+				if (!finite(numval[v][nobs]))
+					{
+					fprintf(dap_err,
+						"(group) NaN value %d for %s\n",
+						nobs, dap_obs[0].do_nam[varv[v]]);
+					exit(1);
+					}
+				}
 			}
 		else
 			{
-			fputs("(group) Too many data.\n", dap_err);
+			fputs("(group) too many data.\n", dap_err);
 			exit(1);
 			}
 		}
 	}
-if (nonum)
+if (nvar > 0)
 	{
 	dap_free(nummem);
 	dap_free(numval);
+	dap_free(ptmem);
+	dap_free(point);
 	}
 dap_free(outname);
 dap_free(grpname);
@@ -952,9 +949,6 @@ dap_free(grpv);
 dap_free(markv);
 dap_free(varv);
 dap_free(ctype);
-dap_free(numv);
-dap_free(ptmem);
-dap_free(point);
 }
 
 #define NFREQSTAT 13
@@ -1104,7 +1098,7 @@ tab[findlev(varv[0], level[0], nlevels)][findlev(varv[1], level[1], nlevels + 1)
 dap_swap();
 }
 
-void valcpy(int v, char *val)
+static void valcpy(int v, char *val)
 {
 if (dap_obs[0].do_len[v] > 0)
 	strcpy(dap_obs[0].do_str[v], val);
@@ -1749,7 +1743,7 @@ dap_free(level[0]);
 dap_free(level[1]);
 }
 
-static void trim1(int *varv, double *vpct, int nvar, double **val,
+static void trim1(double *vpct, int nvar, double **val,
 		int nobs, double *vmin, double *vmax)
 {
 int v;
@@ -1764,7 +1758,7 @@ for (v = 0; v < nvar; v++)
 	}
 }
 
-static int trimparse(char *trimspec, int *varv, double *vpct, char *outlist)
+static int trimparse(char *trimspec, int *varv, double *vpct)
 {
 int n;
 int i;
@@ -1774,7 +1768,6 @@ double div;
 int digits;
 int nvar;
 
-outlist[0] = '\0';
 if (!trimspec)
 	return 0;
 varname = dap_malloc(dap_namelen + 1, "");
@@ -1789,7 +1782,7 @@ for (nvar = 0; trimspec[n]; )
 		else
 			{
 			varname[i] = '\0';
-			fprintf(dap_err, "(trimparse) Trim variable name too long: %s\n",
+			fprintf(dap_err, "(trimparse) trim variable name too long: %s\n",
 				varname);
 			exit(1);
 			}
@@ -1810,7 +1803,7 @@ for (nvar = 0; trimspec[n]; )
 				if (div >= 1.0)
 					{
 					fprintf(dap_err,
-					"(trimparse) Multiple decimal points in percent for %s\n",
+					"(trimparse) multiple decimal points in percent for %s\n",
 						varname);
 					exit(1);
 					}
@@ -1826,8 +1819,6 @@ for (nvar = 0; trimspec[n]; )
 		vpct[nvar] = (double) digits;
 		if (div >= 1.0)
 			vpct[nvar] /= div;
-		strcat(outlist, varname);
-		strcat(outlist, " ");
 		if (vpct[nvar] > 0.0)
 			{
 			if (dap_obs[0].do_len[v] == DBL)
@@ -1835,15 +1826,22 @@ for (nvar = 0; trimspec[n]; )
 			else
 				{
 				fprintf(dap_err,
-					"(trimparse) Trim variable not double: %s\n",
+					"(trimparse) trim variable not double: %s\n",
 						varname);
 				exit(1);
 				}
 			}
+		else
+			{
+			fprintf(dap_err,
+				"(trimparse) no percent for trim variable: %s\n",
+					varname);
+			exit(1);
+			}
 		}
 	else
 		{
-		fprintf(dap_err, "(trimparse) Unknown trim variable: %s\n", varname);
+		fprintf(dap_err, "(trimparse) unknown trim variable: %s\n", varname);
 		exit(1);
 		}
 	while (trimspec[n] == ' ')
@@ -1861,7 +1859,6 @@ int nmark;
 int *varv;
 double *vpct;
 int nvar;
-char *outlist;
 double *valmem;
 double **val;
 int v;
@@ -1877,7 +1874,6 @@ if (!fname)
 	}
 outname = dap_malloc(strlen(fname) + 5, "");
 dap_suffix(outname, fname, ".trm");
-outlist = dap_malloc(dap_listlen + 1, "");
 markv = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
 varv = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
 vpct = (double *) dap_malloc(sizeof(double) * dap_maxvar, "");
@@ -1885,14 +1881,8 @@ vmin = (double *) dap_malloc(sizeof(double) * dap_maxvar, "");
 vmax = (double *) dap_malloc(sizeof(double) * dap_maxvar, "");
 inset(fname);
 nmark = dap_list(marks, markv, dap_maxvar);
-nvar = trimparse(trimspec, varv, vpct, outlist);
-for (v = 0; v < nmark; v++)
-	{
-	strcat(outlist, dap_obs[0].do_nam[markv[v]]);
-	if (v < nmark - 1)
-		strcat(outlist, " ");
-	}
-outset(outname, outlist);
+nvar = trimparse(trimspec, varv, vpct);
+outset(outname, "");
 valmem = (double *) dap_malloc(sizeof(double) * nvar * dap_maxval, "");
 val = (double **) dap_malloc(sizeof(double *) * nvar, "");
 for (v = 0; v <= nvar; v++)
@@ -1902,7 +1892,7 @@ for (dap_mark(), nobs = 0, more = 1; more; nobs++)
 	more = step();
 	if (dap_newpart(markv, nmark))
 		{
-		trim1(varv, vpct, nvar, val, nobs, vmin, vmax);
+		trim1(vpct, nvar, val, nobs, vmin, vmax);
 		dap_rewind();
 		while (step() && !dap_newpart(markv, nmark))
 			{
@@ -1921,7 +1911,17 @@ for (dap_mark(), nobs = 0, more = 1; more; nobs++)
 	if (nobs < dap_maxval)
 		{
 		for (v = 0; v < nvar; v++)
-			val[v][nobs] = dap_obs[0].do_dbl[varv[v]];
+			{
+			if (finite(dap_obs[0].do_dbl[varv[v]]))
+				val[v][nobs] = dap_obs[0].do_dbl[varv[v]];
+			else
+				{
+				fprintf(dap_err,
+					"(trim) NaN value %d for %s\n",
+					nobs, dap_obs[0].do_nam[varv[v]]);
+				exit(1);
+				}
+			}
 		}
 	else
 		{
@@ -1933,7 +1933,6 @@ dap_free(outname);
 dap_free(markv);
 dap_free(varv);
 dap_free(vpct);
-dap_free(outlist);
 dap_free(valmem);
 dap_free(val);
 dap_free(vmin);
