@@ -1,6 +1,6 @@
 /* dap3.c -- plotting routines */
 
-/*  Copyright (C) 2001  Susan Bassein
+/*  Copyright (C) 2001 Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 
 extern dataobs dap_obs[];
 extern FILE *dap_lst;
+extern FILE *dap_log;
 extern FILE *dap_err;
 extern char *dap_title;
 extern char *dap_dapname;
@@ -236,6 +237,7 @@ pict *a;
 int *markv;
 int nmark;
 int nobs;
+int nnan;
 static double *x, *y;
 int pn;
 int xyv[2];
@@ -296,7 +298,7 @@ if (style && style[s] == 'o')
 	if (!overlay)
 		overlay = -1;
 	}
-for (nobs = 0, pn = 0, more = 1; more; nobs++)
+for (nobs = 0, nnan = 0, pn = 0, more = 1; more; )
         {
 	more = step();
         if (dap_newpart(markv, nmark))
@@ -307,6 +309,8 @@ for (nobs = 0, pn = 0, more = 1; more; nobs++)
 				markv, nmark, p, a, pn, style, xfunct, yfunct,
 				(!more || !overlay || (overlay > 0 && !((pn + 1) % overlay))));
 			pn++;
+			if (nnan > 0)
+				fprintf(dap_log, "(plot) %d NaNs\n", nnan);
 			}
 		else
 			{
@@ -316,11 +320,16 @@ for (nobs = 0, pn = 0, more = 1; more; nobs++)
 			exit(1);
 			}
 		nobs = 0;
+		nnan = 0;
 		}
 	if (nobs < dap_maxval)
 		{
 		x[nobs] = dap_obs[0].do_dbl[xyv[0]];
 		y[nobs] = dap_obs[0].do_dbl[xyv[1]];
+		if (finite(x[nobs]) && finite(y[nobs]))
+			nobs++;
+		else
+			nnan++;
 		}
 	else
 		{
@@ -617,6 +626,7 @@ pict *a;
 int *markv;
 int nmark;
 int nobs;
+int nnan;
 double *x, *y;
 int pn;
 int vy;
@@ -682,14 +692,18 @@ if (dap_obs[0].do_len[vy] != DBL)
 	fprintf(dap_err, "(normal) Variable is not double variable: %s\n", varname);
 	exit(1);
 	}
-for (nobs = 0, pn = 0, more = 1; more; nobs++)
+for (nobs = 0, nnan = 0, pn = 0, more = 1; more; )
         {
 	more = step();
         if (dap_newpart(markv, nmark))
 		{
 		if (pn < nplots)
+			{
 			normal1(x, y, nobs, varname, varlabel, markv, nmark,
 				p, l, a, pn++);
+			if (nnan > 0)
+				fprintf(dap_log, "(normal) %d NaNs\n", nnan);
+			}
 		else
 			{
 			fprintf(dap_err,
@@ -698,9 +712,16 @@ for (nobs = 0, pn = 0, more = 1; more; nobs++)
 			exit(1);
 			}
 		nobs = 0;
+		nnan = 0;
 		}
 	if (nobs < dap_maxval)
+		{
 		y[nobs] = dap_obs[0].do_dbl[vy];
+		if (finite(y[nobs]))
+			nobs++;
+		else
+			nnan++;
+		}
 	else
 		{
 		fprintf(dap_err, "(normal) Too many points\n");
@@ -744,7 +765,7 @@ double *part;
 int equal;
 int height;
 int whole;
-int zero;
+double min, max;
 double xlen, xspace;
 static char htitle[19];
 double width;
@@ -769,9 +790,20 @@ if (!nbars)
 equal = 1;
 height = UNSPEC;
 whole = 0;
-zero = 0;
 htitle[0] = '\0';
 axspec[0] = '\0';
+if (x)
+	{
+	qsort((void *) x, (size_t) nobs, (size_t) sizeof(double), cmp);
+	part[0] = x[0];
+	part[nbars] = x[nobs - 1];
+	}
+else
+	{
+	qsort((void *) xw, (size_t) nobs, (size_t) (2 * sizeof(double)), cmp);
+	part[0] = xw[0][0];
+	part[nbars] = xw[nobs - 1][0];
+	}
 if (style)
 	{
 	for (s = 0; style[s] == ' '; s++)
@@ -802,8 +834,26 @@ if (style)
 			height = FRACTION;
 		else if (!strcmp(word, "ROUND"))
 			whole = 1;
-		else if (!strcmp(word, "ZERO"))
-			zero = 1;
+		else if (!strncmp(word, "MIN", 3))
+			{
+			if (sscanf(word + 3, "%lf", &part[0]) != 1)
+				{
+				fprintf(dap_err,
+					"(histo1) bad MIN specification: %s\n",
+					word);
+				exit(1);
+				}
+			}
+		else if (!strncmp(word, "MAX", 3))
+			{
+			if (sscanf(word + 3, "%lf", &part[nbars]) != 1)
+				{
+				fprintf(dap_err,
+					"(histo1) bad MAX specification: %s\n",
+					word);
+				exit(1);
+				}
+			}
 		else if (index("-+0nb", word[0]))
 			{
 			strncpy(axspec, word, 2);
@@ -851,22 +901,8 @@ else
 		break;
 		}
 	}
-if (x)
-	{
-	qsort((void *) x, (size_t) nobs, (size_t) sizeof(double), cmp);
-	part[0] = x[0];
-	part[nbars] = x[nobs - 1];
-	}
-else
-	{
-	qsort((void *) xw, (size_t) nobs, (size_t) (2 * sizeof(double)), cmp);
-	part[0] = xw[0][0];
-	part[nbars] = xw[nobs - 1][0];
-	}
 dnobs = (double) nobs;
 dnbars = (double) nbars;
-if (zero)
-	part[0] = 0.0;
 if (whole)
 	{
 	xlen = 1e5 / (part[nbars] - part[0]);
@@ -1044,20 +1080,20 @@ for (nobs = 0, nnan = 0, pn = 0, more = 1; more; )
                 dap_swap();
 		if (nnan)
 			{
-			fprintf(dap_err, "(histogram) %d missing values for:", nnan);
+			fprintf(dap_log, "(histogram) %d missing values for:", nnan);
 			for (mv = 0; mv < nmark; mv++)
 				{
-				putc(' ', dap_err);
+				putc(' ', dap_log);
 				if (dap_obs[0].do_len[markv[mv]] == DBL)
-					fprintf(dap_err, "%g",
+					fprintf(dap_log, "%g",
 						dap_obs[0].do_dbl[markv[mv]]);
 				else if (dap_obs[0].do_len[markv[mv]] == INT)
-					fprintf(dap_err, "%d",
+					fprintf(dap_log, "%d",
 						dap_obs[0].do_int[markv[mv]]);
 				else
-					fputs(dap_obs[0].do_str[markv[mv]], dap_err);
+					fputs(dap_obs[0].do_str[markv[mv]], dap_log);
 				}
-			putc('\n', dap_err);
+			putc('\n', dap_log);
 			}
 		if (nobs)
 			{
@@ -1374,7 +1410,7 @@ if (noverlay < 1)
 	noverlay = 1;
 sprintf(overstr, "o%d bb", 2 * noverlay);
 srtarg = dap_malloc(strlen(mnslist) + 8, "");
-if (partvars)
+if (partvars && partvars[0])
 	{
 	strcpy(srtarg, partvars);
 	strcat(srtarg, " ");
@@ -1410,13 +1446,13 @@ inset(mnsname);
 outset(errname, "");
 dap_list(varlist, meanv, 1);	/* check that there's only one */
 dap_list(meanvar, meanv, 1);
-if (partvars)
+if (partvars && partvars[0])
 	npartv = dap_list(partvars, partv, dap_maxvar);
 else
 	npartv = 0;
 if ((typen = dap_varnum("_type_")) < 0)
 	{
-	fprintf(stderr, "%s: missing _type_ variable\n", dap_dapname);
+	fprintf(stderr, "%s: missing _type_ variable for plotmeans\n", dap_dapname);
 	exit(1);
 	}
 for (n = 0, nparts = 0, more = 1; more; )
