@@ -1,6 +1,6 @@
 /* dap5.c -- regression and nonparametric statistics */
 
-/*  Copyright (C) 2001  Susan Bassein
+/*  Copyright (C) 2001 Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -618,6 +618,9 @@ double *v;
 double *step;
 double loglike0, loglike1;
 int niter;
+double maxv;
+double maxcov;
+double tmp;
 
 beta1 = (double *) dap_malloc(sizeof(double) * nx, "");
 v = (double *) dap_malloc(sizeof(double) * nx, "");
@@ -635,16 +638,30 @@ do	{
 	for (i = 0; i < nx; i++)
 		beta0[i] = beta1[i];
 	loglike0 = loglike1;
-	for (i = 0; i < nx; i++)
+	for (i = 0, maxv = 0.0, maxcov = 0.0; i < nx; i++)
 		{
 		for (v[i] = 0.0, n = 0; n < nobs; n++)
 			v[i] += x[i][n] * (y[0][n] - y[1][n] * pr[n]);
+		if ((tmp = fabs(v[i])) > maxv)
+			maxv = tmp;
 		for (j = 0; j < nx; j++)
 			{
 			cov[i][j] = 0.0;
 			for (n = 0; n < nobs; n++)
 				cov[i][j] += y[1][n] * pr[n] * (1.0 - pr[n]) *
 								x[i][n] * x[j][n];
+			if ((tmp = fabs(cov[i][j])) > maxcov)
+				maxcov = tmp;
+			}
+		}
+	for (i = 0; i < nx; i++)
+		{
+		if (fabs(v[i]) < dap_ctol * maxv)
+			v[i] = 0.0;
+		for (j = 0; j < nx; j++)
+			{
+			if (fabs(cov[i][j]) < dap_ctol * maxcov)
+				cov[i][j] = 0.0;
 			}
 		}
 	if (!dap_invert(cov, nx))
